/* SPDX-License-Identifier: LGPL-2.1+ */
/******************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2020. Allrights reserved
 * Description: isulad utils
 * Author: lifeng
 * Create: 2020-04-11
******************************************************************************/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio_ext.h>

#include "isulad_utils.h"
#include "log.h"
#include "path.h"
#include "file_utils.h"

lxc_log_define(isulad_utils, lxc);

void *lxc_common_calloc_s(size_t size)
{
	if (size == 0 || size > SIZE_MAX) {
		return NULL;
	}

	return calloc((size_t)1, size);
}

int lxc_mem_realloc(void **newptr, size_t newsize, void *oldptr, size_t oldsize)
{
	void *tmp = NULL;

	if (newsize == 0) {
		goto err_out;
	}

	tmp = lxc_common_calloc_s(newsize);
	if (tmp == NULL) {
		ERROR("Failed to malloc memory");
		goto err_out;
	}

	if (oldptr != NULL) {
		memcpy(tmp, oldptr, (newsize < oldsize) ? newsize : oldsize);

		memset(oldptr, 0, oldsize);

		free(oldptr);
	}

	*newptr = tmp;
	return 0;

err_out:
	return -1;
}

char *safe_strdup(const char *src)
{
	char *dst = NULL;

	if (src == NULL) {
		return NULL;
	}

	dst = strdup(src);
	if (dst == NULL) {
		abort();
	}

	return dst;
}

int lxc_open(const char *filename, int flags, mode_t mode)
{
	char rpath[PATH_MAX] = {0x00};

	if (cleanpath(filename, rpath, sizeof(rpath)) == NULL) {
		return -1;
	}
	if (mode) {
		return open(rpath, (int)((unsigned int)flags | O_CLOEXEC), mode);
	} else {
		return open(rpath, (int)((unsigned int)flags | O_CLOEXEC));
	}
}

FILE *lxc_fopen(const char *filename, const char *mode)
{
	char rpath[PATH_MAX] = {0x00};

	if (cleanpath(filename, rpath, sizeof(rpath)) == NULL) {
		return NULL;
	}

	return fopen_cloexec(rpath, mode);
}

static int hold_int(const char delim, bool required, char **src, unsigned int *dst)
{
    unsigned long long int res = 0;
    char *err_str = NULL;

    // ensure *src not a empty string
    if (**src == '\0') {
        ERROR("Empty subject on given entrie is not allowed.");
        return -1;
    }

    errno = 0;
    // covert string to long long
    res = strtoull(*src, &err_str, 0);
    if (errno != 0 && errno != ERANGE) {
        ERROR("Parse int from string failed.");
        return -1;
    }

    // **src is not a digit
    if (err_str == *src) {
        if (!required) {
            ERROR("Integer part is missing.");
            return -1;
        }
        // if required, just set 0
        *dst = 0;
    } else {
        if (sizeof(void *) > 4 && res > UINT_MAX) { // make sure 64-bit platform behave same as 32-bit
            res = UINT_MAX;
        }
        res = res & UINT_MAX;
        *dst = (uint32_t)res;
    }

    // normal case
    if (*err_str == delim) {
        err_str++;
    } else if (*err_str != '\0') {
        ERROR("Invalid digit string.");
        return -1;
    }

    *src = err_str; // update src to next valid context in line.
    return 0;
}

static void hold_string(const char delim, char **src, char **dst)
{
    for (*dst = *src; **src != delim; ++(*src)) {
        if (**src == '\0') {
            break;
        }
    }

    if (**src == delim) {
        **src = '\0';
        ++(*src);
    }
}

static int parse_line_pw(const char delim, char *line, struct passwd *result)
{
    int ret = 0;
    bool required = false;
    char *walker = NULL;

    walker = strpbrk(line, "\n");
    if (walker != NULL) {
        // clear newline char
        *walker = '\0';
    }

    hold_string(delim, &line, &result->pw_name);

    required = (result->pw_name[0] == '+' || result->pw_name[0] == '-') ? true : false;

    hold_string(delim, &line, &result->pw_passwd);

    ret = hold_int(delim, required, &line, &result->pw_uid);
    if (ret != 0) {
        // a legitimate line must have uid
        ERROR("Parse uid error.");
        return ret;
    }

    ret = hold_int(delim, required, &line, &result->pw_gid);
    if (ret != 0) {
        // it's ok to not provide gid
        ERROR("Parse gid error.");
        return ret;
    }

    hold_string(delim, &line, &result->pw_gecos);

    hold_string(delim, &line, &result->pw_dir);

    result->pw_shell = line;
    return 0;
}

char *util_left_trim_space(char *str)
{
    char *begin = str;
    char *tmp = str;
    while (isspace(*begin)) {
        begin++;
    }
    while ((*tmp++ = *begin++)) {
    }
    return str;
}

int util_getpwent_r(FILE *stream, struct passwd *resbuf, char *buffer, size_t buflen, struct passwd **result)
{
    const char delim = ':';
    char *buff_end = NULL;
    char *walker = NULL;
    bool got = false;
    int ret = 0;

    if (stream == NULL || resbuf == NULL || buffer == NULL || result == NULL) {
        ERROR("Password obj, params is NULL.");
        return -1;
    }

    if (buflen <= 1) {
        ERROR("Inadequate buffer length was given.");
        return -1;
    }

    buff_end = buffer + buflen - 1;
    flockfile(stream);

    while (1) {
        *buff_end = '\xff';
        walker = fgets_unlocked(buffer, buflen, stream);
        // if get NULL string
        if (walker == NULL) {
            *result = NULL;
            // reach end of file, return error
            if (feof(stream)) {
                ret = ENOENT;
                goto out;
            }
            // overflow buffer
            ret = ERANGE;
            goto out;
        }
        // just overflow last char in buffer
        if (*buff_end != '\xff') {
            *result = NULL;
            ret = ERANGE;
            goto out;
        }

        (void)util_left_trim_space(buffer);
        // skip comment line and empty line
        if (walker[0] == '#' || walker[0] == '\0') {
            continue;
        }

        if (parse_line_pw(delim, walker, resbuf) == 0) {
            got = true;
            break;
        }
    }
    if (!got) {
        *result = NULL;
        ret = ERANGE;
        goto out;
    }

    *result = resbuf;
    ret = 0;
out:
    funlockfile(stream);
    return ret;
}